/*TITLE conversions to/from Radix40 representation */

/****keyword-flag*** "%v %f %n" */
/* "4 20-Mar-98,20:14:08 RADIX40.CPP" */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <time.h>
#include <stdlib.h>
#include "radix40.h"
#include <windows.h>

#if _DEBUG
#define NUMBER_LOOPS		1
#else
#define NUMBER_LOOPS		100000
#endif // #if _DEBUG

#define MAX_CHARS			100

#define USE_ASM				0
#if USE_ASM
#define	ALGO1				1
#define	ALGO2				2
#define	ALGO3				3
#define	ALGO4				4
#define USE_ALGO			ALGO4
#endif

// #define C1 = original
// #define C2 = avoids using % to cycle 0,1,2
// #define C3 = using pointers instead of indexing
// #define C4 = unrolled to avoid cycle and accessing weight[]
// #define C5 = better loop unrolling

int ascii_to_radix40(Radix40 *radix40_data, unsigned char *ascii_data, int max_chars);
int ascii_to_radix40_mark(Radix40 *radix40_data, unsigned char *ascii_data, int max_chars);
void print_array(Radix40 *radix40_data);


char legal_chars[41] = " ,-.0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";

unsigned weights[3] = {1600, 40, 1};

unsigned int table1[40], table40[40], table1600[40];

//all structures are 16 byte aligned.
Radix40 radix40_arr[MAX_CHARS];
unsigned char ascii_arr[MAX_CHARS+1];
long counter;

__int64 timer1;
double duration, base_duration;

void _cdecl main()
{
//set process priority
	HANDLE cur_proc = GetCurrentProcess();
	SetPriorityClass( cur_proc, REALTIME_PRIORITY_CLASS);
	srand(2);
	for ( int t = 0; t < 100; t++)
		ascii_arr[t] = legal_chars[rand() % 40];

for (int t = 0; t < 40; t++)
{
    table1[t] = t;
    table40[t] = t*40;
    table1600[t] = t*1600;
}

	for ( t = 0; t < NUMBER_LOOPS; t++)
	{

		_asm {
			cpuid
			rdtsc
			mov		dword ptr [timer1],eax
			mov		dword ptr [timer1+4],edx
		}
//#if 0
#if _DEBUG
			ascii_to_radix40_mark(radix40_arr, ascii_arr, MAX_CHARS);
			print_array(radix40_arr);
			ascii_to_radix40(radix40_arr, ascii_arr, MAX_CHARS);
			print_array(radix40_arr);
#else
#if USE_ASM
			ascii_to_radix40_mark(radix40_arr, ascii_arr, MAX_CHARS);
#else
			ascii_to_radix40(radix40_arr, ascii_arr, MAX_CHARS);
#endif
#endif // #if _DEBUG
//#endif // #if 0
		_asm {
			rdtsc
			sub		eax,dword ptr [timer1]
			sbb		edx,dword ptr [timer1+4]
			sub		eax,84						//P4 takes 84 cycles to run this loop empty
			sbb		edx,0
			mov		dword ptr [timer1],eax
			mov		dword ptr [timer1+4],edx
			emms
		}
//		duration = (double)timer1;
		printf("%.6I64d cycles\n", timer1);
	}

	SetPriorityClass( cur_proc, NORMAL_PRIORITY_CLASS);

}

void print_array(Radix40 *radix40_data)
{
	int t;

	for ( t=0; t<MAX_CHARS; t++)
	{
		if ( radix40_data[t] == 0)
			break;
		printf("%.8LX ", radix40_data[t]);
	}
	printf("\n\n");
}

char lookup_chars[256] =
{IL, IL, IL, IL, IL, IL, IL, IL,/* 00 */
IL, IL, IL, IL, IL, IL, IL, IL, /* 08 */
IL, IL, IL, IL, IL, IL, IL, IL, /* 10 */
IL, IL, IL, IL, IL, IL, IL, IL, /* 18 */
 0, IL, IL, IL, IL, IL, IL, IL, /* 20 */
IL, IL, IL, IL,  1,  2,  3, IL, /* 28 */
 4,  5,  6,  7,  8,  9, 10, 11, /* 30 */
12, 13 ,IL, IL, IL, IL, IL, IL, /* 38 */
IL, 14, 15, 16, 17, 18, 19, 20, /* 40 */
21, 22, 23, 24, 25, 26, 27, 28, /* 48 */
29, 30, 31, 32, 33, 34, 35, 36, /* 50 */
37, 38, 39, IL, IL, IL, IL, IL, /* 58 */
IL, 14, 15, 16, 17, 18, 19, 20, /* 60 */
21, 22, 23, 24, 25, 26, 27, 28, /* 68 */
29, 30, 31, 32, 33, 34, 35, 36, /* 70 */
37, 38, 39, IL, IL, IL, IL, IL, /* 78 */
IL, IL, IL, IL, IL, IL, IL, IL, /* 80 */
IL, IL, IL, IL, IL, IL, IL, IL, /* 88 */
IL, IL, IL, IL, IL, IL, IL, IL, /* 90 */
IL, IL, IL, IL, IL, IL, IL, IL, /* 98 */
IL, IL, IL, IL, IL, IL, IL, IL, /* A0 */
IL, IL, IL, IL, IL, IL, IL, IL, /* A8 */
IL, IL, IL, IL, IL, IL, IL, IL, /* B0 */
IL, IL, IL, IL, IL, IL, IL, IL, /* B8 */
IL, IL, IL, IL, IL, IL, IL, IL, /* C0 */
IL, IL, IL, IL, IL, IL, IL, IL, /* C8 */
IL, IL, IL, IL, IL, IL, IL, IL, /* D0 */
IL, IL, IL, IL, IL, IL, IL, IL, /* D8 */
IL, IL, IL, IL, IL, IL, IL, IL, /* E0 */
IL, IL, IL, IL, IL, IL, IL, IL, /* E8 */
IL, IL, IL, IL, IL, IL, IL, IL, /* F0 */
IL, IL, IL, IL, IL, IL, IL, IL  /* F8 */
};

//best time ? cycles
__forceinline int ascii_to_radix40(Radix40 *radix40_data, unsigned char *ascii_data, int max_chars)
//int __cdecl ascii_to_radix40(Radix40 *radix40_data, unsigned char *ascii_data, int max_chars)
/* this routine converts a null-terminated ascii character string */
/* to radix 40 representation.  The allowable characters are: */
/* A-Z, 0-9, period, comma, hyphen, and space.  If any illegal characters */
/* are detected, they will be converted to hyphens, and the return value */
/* will be S_ILLEGAL. */
/* Lowercase letters will be upper-cased without error indication. */
/* The radix40 value will be padded with blanks to a multiple of three */
/* characters. If the number of characters in the ascii string is > max_chars */
/* only max_chars will be converted, and the return value will be S_ILLEGAL. */
/* If no error is detected, the return value will be S_OKAY. */
{
	int ascii_length;
	int result;
	int words_to_convert;
	int words_to_clear;
#if defined(C1) || defined(C2)
	int i;
	unsigned char j;
	int cycle;
	unsigned current_word_index;
#endif

	result = S_OKAY;
	ascii_length = strlen((char *)ascii_data);
	if (ascii_length > max_chars)
		{
		ascii_length = max_chars;
		result = S_ILLEGAL;
		}

	words_to_convert = ascii_length / 3;
	if (ascii_length % 3 != 0)
		words_to_convert ++;

	words_to_clear = max_chars / 3;
	if (max_chars % 3 != 0)
		words_to_clear ++;

	memset(radix40_data,0,words_to_clear*sizeof(Radix40));

// C1 is the original, C2 avoids using % to cycle 0,1,2
#if defined(C1) || defined(C2)
	current_word_index = unsigned(-1);
	cycle = 0;
	for (i = 0; i < ascii_length; i ++)
		{
		if (cycle == 0)
			current_word_index ++;
		j = lookup_chars[ascii_data[i]];
		if (j & ILLEGAL)
			{
			j = HYPHEN ; /* make it a hyphen */
			result = S_ILLEGAL; /* and remember that it was illegal */
			}
		radix40_data[current_word_index] += weights[cycle] * j;
#ifdef C1
		cycle = (cycle + 1) % 3;
#else
		if (++cycle == 3) cycle = 0;
#endif
		}
#endif

// using pointers instead of indexing
#ifdef C3
	int cycle = 0;
	for (unsigned char * const limit = ascii_data + ascii_length; ascii_data < limit; )
		{
		unsigned int j = (unsigned char)lookup_chars[*ascii_data++];
		if (j & ILLEGAL)
			{
			j = HYPHEN ; /* make it a hyphen */
			result = S_ILLEGAL; /* and remember that it was illegal */
			}
		*radix40_data += weights[cycle] * j;
		if (++cycle == 3) { cycle = 0; ++radix40_data; }
		}
#endif

// unrolled to avoid cycle and accessing weight[]
#ifdef C4
	for (unsigned char * const limit = ascii_data + ascii_length; ascii_data < limit; )
		{
		unsigned int j = (unsigned char)lookup_chars[*ascii_data];
		if (j & ILLEGAL)
			{
			j = HYPHEN ; /* make it a hyphen */
			result = S_ILLEGAL; /* and remember that it was illegal */
			}
		*radix40_data += 1600 * j;
		if (++ascii_data >= limit) break;
		j = (unsigned char)lookup_chars[*ascii_data];
		if (j & ILLEGAL)
			{
			j = HYPHEN ; /* make it a hyphen */
			result = S_ILLEGAL; /* and remember that it was illegal */
			}
		*radix40_data += 40 * j;
		if (++ascii_data >= limit) break;
		j = (unsigned char)lookup_chars[*ascii_data];
		if (j & ILLEGAL)
			{
			j = HYPHEN ; /* make it a hyphen */
			result = S_ILLEGAL; /* and remember that it was illegal */
			}
		*radix40_data += j;
		if (++ascii_data >= limit) break;
		++radix40_data;
		}
#endif

// better loop unrolling
#ifdef C5
while (ascii_length >= 3)
		{
		unsigned int j = (unsigned char)lookup_chars[ascii_data[0]];
		if (j & ILLEGAL)
			{
			j = HYPHEN ; /* make it a hyphen */
			result = S_ILLEGAL; /* and remember that it was illegal */
			}
		*radix40_data += 1600 * j;
		j = (unsigned char)lookup_chars[ascii_data[1]];
		if (j & ILLEGAL)
			{
			j = HYPHEN ; /* make it a hyphen */
			result = S_ILLEGAL; /* and remember that it was illegal */
			}
		*radix40_data += 40 * j;
		j = (unsigned char)lookup_chars[ascii_data[2]];
		if (j & ILLEGAL)
			{
			j = HYPHEN ; /* make it a hyphen */
			result = S_ILLEGAL; /* and remember that it was illegal */
			}
		*radix40_data++ += j;
		ascii_data += 3;
		ascii_length -= 3;
		}

switch (ascii_length)
{
case 2:
{
		unsigned int j = (unsigned char)lookup_chars[ascii_data[0]];
		if (j & ILLEGAL)
			{
			j = HYPHEN ; /* make it a hyphen */
			result = S_ILLEGAL; /* and remember that it was illegal */
			}
		*radix40_data += 1600 * j;
		j = (unsigned char)lookup_chars[ascii_data[1]];
		if (j & ILLEGAL)
			{
			j = HYPHEN ; /* make it a hyphen */
			result = S_ILLEGAL; /* and remember that it was illegal */
			}
		*radix40_data += 40 * j;
}
break;
case 1:
{
		unsigned int j = (unsigned char)lookup_chars[ascii_data[0]];
		if (j & ILLEGAL)
			{
			j = HYPHEN ; /* make it a hyphen */
			result = S_ILLEGAL; /* and remember that it was illegal */
			}
		*radix40_data += 1600 * j;
}
break;
}
#endif

	return(result);
}

#if USE_ASM
//best time 2496 cycles
#if USE_ALGO==ALGO1
__forceinline int ascii_to_radix40_mark(Radix40 *radix40_data, unsigned char *ascii_data, int max_chars)
/* this routine converts a null-terminated ascii character string */
/* to radix 40 representation.  The allowable characters are: */
/* A-Z, 0-9, period, comma, hyphen, and space.  If any illegal characters */
/* are detected, they will be converted to hyphens, and the return value */
/* will be S_ILLEGAL. */
/* Lowercase letters will be upper-cased without error indication. */
/* The radix40 value will be padded with blanks to a multiple of three */
/* characters. If the number of characters in the ascii string is > max_chars */
/* only max_chars will be converted, and the return value will be S_ILLEGAL. */
/* If no error is detected, the return value will be S_OKAY. */
{
	int i;
	unsigned char j;
	int ascii_length;
	int result;
	int words_to_convert;
	int words_to_clear;
	int cycle;
	unsigned current_word_index;

	result = S_OKAY;
	ascii_length = strlen((char *)ascii_data);
	if (ascii_length > max_chars)
		{
		ascii_length = max_chars;
		result = S_ILLEGAL;
		}

	words_to_convert = ascii_length / 3;
	if (ascii_length % 3 != 0)
		words_to_convert ++;

	words_to_clear = max_chars / 3;
	if (max_chars % 3 != 0)
		words_to_clear ++;


	_asm {
		push ebx
		push esi
		push edi
//	memset(radix40_data,0,words_to_clear*sizeof(Radix40));
		mov		ecx,[words_to_clear]			//Get number of bytes to clear.
		xor		eax,eax
		mov		edi,[radix40_data]				//load up EDI with offset
		rep		stosd

		mov		eax,[ascii_length]				//eax = i = ascii_length ( we are counting down instead of up)
		xor		edx,edx							//cycle = edx = 0
		mov		ebx,-1							//current_word_index = ebx = -1
		mov		esi,[radix40_data]
		mov		edi,[ascii_data]
align 4
top_of_loop:
		test	edx,edx							//cycle == 0?
		jnz		skip_increment
		inc		ebx								//current_word_index++
skip_increment:
		movzx	ecx,byte ptr [edi]				//ecx = ascii_data
		movzx	ecx,byte ptr [lookup_chars+ecx]			//j = lookup_chars[ascii_data[i]];
		test	cl,ILLEGAL
		jz		not_illegal
		
		movzx	ecx,HYPHEN						//j = HYPHEN
		mov		dword ptr [result],1			//S_ILLEGAL was not working.
not_illegal:
		push	eax								//radix40_data[current_word_index] += weights[cycle] * j;
		push	edx
		mov		eax,[weights+edx*4]				//edx = weights[cycle]
		mul		ecx								//eax:edx = weights[cycle] * j;
		add		[esi+ebx*4],eax					//radix40_data[current_word_index] += weights[cycle] * j;
		pop		edx
		pop		eax

		inc		edx								//cycle = (cycle + 1) % 3;
		cmp		edx,3
		jb		skip_it
		xor		edx,edx							//cycle = 0
skip_it:

		inc		edi								//point to next array location in the ASCII array.

		dec		eax
		jnz		top_of_loop

		pop edi
		pop esi
		pop ebx
	}
	return(result);
}
#endif // #if USE_ALGO==ALGO1

//best time is 2028 cycles, set real time priority class
#if USE_ALGO==ALGO2
__forceinline int ascii_to_radix40_mark(Radix40 *radix40_data, unsigned char *ascii_data, int max_chars)
/* this routine converts a null-terminated ascii character string */
/* to radix 40 representation.  The allowable characters are: */
/* A-Z, 0-9, period, comma, hyphen, and space.  If any illegal characters */
/* are detected, they will be converted to hyphens, and the return value */
/* will be S_ILLEGAL. */
/* Lowercase letters will be upper-cased without error indication. */
/* The radix40 value will be padded with blanks to a multiple of three */
/* characters. If the number of characters in the ascii string is > max_chars */
/* only max_chars will be converted, and the return value will be S_ILLEGAL. */
/* If no error is detected, the return value will be S_OKAY. */
{
	int i;
	unsigned char j;
	int ascii_length;
	int result;
	int words_to_convert;
	int words_to_clear;
	int cycle;
	unsigned current_word_index;

	result = S_OKAY;
	ascii_length = strlen((char *)ascii_data);
	if (ascii_length > max_chars)
		{
		ascii_length = max_chars;
		result = S_ILLEGAL;
		}

	words_to_convert = ascii_length / 3;
	if (ascii_length % 3 != 0)
		words_to_convert ++;

	words_to_clear = max_chars / 3;
	if (max_chars % 3 != 0)
		words_to_clear ++;


	_asm {
		push ebx
		push esi
		push edi
//	memset(radix40_data,0,words_to_clear*sizeof(Radix40));
		mov		ecx,[words_to_clear]			//Get number of bytes to clear.
		xor		eax,eax
		mov		edi,[radix40_data]				//load up EDI with offset
		rep		stosd

//edx now free since we unrolled it 3 times.
		mov		eax,[ascii_length]				//eax = i = ascii_length ( we are counting down instead of up)
		mov		ebx,-1							//current_word_index = ebx = -1
		mov		esi,[radix40_data]
		mov		edi,[ascii_data]
align 4
top_of_loop:
		inc		ebx								//current_word_index++

		movzx	edx,byte ptr [edi+0]			//edx = ascii_data
		movzx	edx,byte ptr [lookup_chars+edx]	//j = lookup_chars[ascii_data[i]];
		test	dl,ILLEGAL
		jz		not_illegal1
		
		movzx	edx,HYPHEN						//j = HYPHEN
		mov		dword ptr [result],1			//S_ILLEGAL was not working.
not_illegal1:
		push	eax								//radix40_data[current_word_index] += weights[cycle] * j;
mov		eax,[weights+0]					//eax = weights[cycle]
//mov		eax,1600					//eax = weights[cycle]
		mul		edx								//eax:edx = weights[cycle] * j;
		add		[esi+ebx*4],eax					//radix40_data[current_word_index] += weights[cycle] * j;
		pop		eax

		cmp		eax,1
		jbe		end_of_loop

		movzx	edx,byte ptr [edi+1]			//edx = ascii_data
		movzx	edx,byte ptr [lookup_chars+edx]	//j = lookup_chars[ascii_data[i]];
		test	dl,ILLEGAL
		jz		not_illegal2
		
		movzx	edx,HYPHEN						//j = HYPHEN
		mov		dword ptr [result],1			//S_ILLEGAL was not working.
not_illegal2:
		push	eax								//radix40_data[current_word_index] += weights[cycle] * j;
mov		eax,[weights+4]					//eax = weights[cycle]
//mov		eax,40					//eax = weights[cycle]
		mul		edx								//eax:edx = weights[cycle] * j;
		add		[esi+ebx*4],eax					//radix40_data[current_word_index] += weights[cycle] * j;
		pop		eax

		cmp		eax,2
		jbe		end_of_loop

		movzx	edx,byte ptr [edi+2]			//edx = ascii_data
		movzx	edx,byte ptr [lookup_chars+edx]	//j = lookup_chars[ascii_data[i]];
		test	dl,ILLEGAL
		jz		not_illegal3
		
		movzx	edx,HYPHEN						//j = HYPHEN
		mov		dword ptr [result],1			//S_ILLEGAL was not working.
not_illegal3:
		push	eax								//radix40_data[current_word_index] += weights[cycle] * j;
mov		eax,[weights+8]					//eax = weights[cycle]
//mov		eax,1					//eax = weights[cycle]
		mul		edx								//eax:edx = weights[cycle] * j;
		add		[esi+ebx*4],eax					//radix40_data[current_word_index] += weights[cycle] * j;
		pop		eax

		add		edi,3							//point to next array location in the ASCII array.

		sub		eax,3
		jnz		top_of_loop

end_of_loop:

		pop edi
		pop esi
		pop ebx
	}
	return(result);
}
#endif // #if USE_ALGO==ALGO2


//best time 1900 cycles
#if USE_ALGO==ALGO3
__forceinline int ascii_to_radix40_mark(Radix40 *radix40_data, unsigned char *ascii_data, int max_chars)
/* this routine converts a null-terminated ascii character string */
/* to radix 40 representation.  The allowable characters are: */
/* A-Z, 0-9, period, comma, hyphen, and space.  If any illegal characters */
/* are detected, they will be converted to hyphens, and the return value */
/* will be S_ILLEGAL. */
/* Lowercase letters will be upper-cased without error indication. */
/* The radix40 value will be padded with blanks to a multiple of three */
/* characters. If the number of characters in the ascii string is > max_chars */
/* only max_chars will be converted, and the return value will be S_ILLEGAL. */
/* If no error is detected, the return value will be S_OKAY. */
{
	int i;
	unsigned char j;
	int ascii_length;
	int result;
	int words_to_convert;
	int words_to_clear;
	int cycle;
	unsigned current_word_index;

	result = S_OKAY;
	ascii_length = strlen((char *)ascii_data);
	if (ascii_length > max_chars)
		{
		ascii_length = max_chars;
		result = S_ILLEGAL;
		}

	words_to_convert = ascii_length / 3;
	if (ascii_length % 3 != 0)
		words_to_convert ++;

	words_to_clear = max_chars / 3;
	if (max_chars % 3 != 0)
		words_to_clear ++;


	_asm {
		push ebx
		push esi
		push edi
movd mm0,esp		//you cannot push and pop if this is not pointing to the right location.
//	memset(radix40_data,0,words_to_clear*sizeof(Radix40));
		mov		ecx,[words_to_clear]			//Get number of bytes to clear.
		xor		eax,eax
		mov		edi,[radix40_data]				//load up EDI with offset
		rep		stosd

//edx now free since we unrolled it 3 times.
		mov		esp,[ascii_length]				//esp = i = ascii_length ( we are counting down instead of up)
		mov		ebx,-1							//current_word_index = ebx = -1
		mov		esi,[radix40_data]
		mov		edi,[ascii_data]
align 4
top_of_loop:
		inc		ebx								//current_word_index++

mov edx,0		//eax % 3
mov eax,0
cmp esp,1
je do_single

		movzx	edx,byte ptr [edi+0]			//edx = ascii_data
movzx	ecx,byte ptr [edi+1]			//edx = ascii_data

//movzx eax,HYPHEN
		movzx	edx,byte ptr [lookup_chars+edx]	//j = lookup_chars[ascii_data[i]];
movzx	ecx,byte ptr [lookup_chars+ecx]	//j = lookup_chars[ascii_data[i]];

/*
		test	dl,ILLEGAL
//jz		not_illegal1
		
//movzx	edx,HYPHEN						//j = HYPHEN
//mov		dword ptr [result],1			//S_ILLEGAL was not working.
cmovz edx,eax
//cmovz [result],eax  //gets set to 2,  Any non-zero value indicates error
test	cl,ILLEGAL
cmovz ecx,eax
//cmovz [result],eax  //gets set to 2,  Any non-zero value indicates error
*/
mov	eax,edx
and dl,not(0x80)  // do a "hyphen"
or	eax,ecx
and cl,not(0x80)  // do a "hyphen"
test al,ILLEGAL
//GET RID OF THE JUMP BY DOING THIS, and AL,ILLEGAL, OR DWORD PTR [result],EAX
jz not_illegal1
mov		dword ptr [result],1			//S_ILLEGAL was not working.
not_illegal1:
//push	eax								//radix40_data[current_word_index] += weights[cycle] * j;
		mov		eax,1600					//edx = weights[cycle]
		mul		edx								//eax:edx = weights[cycle] * j;
		add		[esi+ebx*4],eax					//radix40_data[current_word_index] += weights[cycle] * j;
mov		eax,40					//edx = weights[cycle]
mul		edx								//eax:edx = weights[cycle] * j;
add		[esi+ebx*4],eax					//radix40_data[current_word_index] += weights[cycle] * j;
//pop		eax

mov edx,2		//eax % 3
mov eax,2
cmp esp,3
je do_single
cmp		esp,2
jbe		end_of_loop





		movzx	edx,byte ptr [edi+2]			//edx = ascii_data
movzx	ecx,byte ptr [edi+3]			//edx = ascii_data

movzx eax,HYPHEN
		movzx	edx,byte ptr [lookup_chars+edx]	//j = lookup_chars[ascii_data[i]];
movzx	ecx,byte ptr [lookup_chars+ecx]	//j = lookup_chars[ascii_data[i]];
/*
		test	dl,ILLEGAL
//jz		not_illegal1
		
//movzx	edx,HYPHEN						//j = HYPHEN
//mov		dword ptr [result],1			//S_ILLEGAL was not working.
cmovz edx,eax
//cmovz [result],eax  //gets set to 2,  Any non-zero value indicates error
test	cl,ILLEGAL
cmovz ecx,eax
//cmovz [result],eax  //gets set to 2,  Any non-zero value indicates error
*/
mov	eax,edx
and dl,not(0x80)  // do a "hyphen"
or	eax,ecx
and cl,not(0x80)  // do a "hyphen"
test al,ILLEGAL
//GET RID OF THE JUMP BY DOING THIS, and AL,ILLEGAL, OR DWORD PTR [result],EAX
jz not_illegal2
mov		dword ptr [result],1			//S_ILLEGAL was not working.
not_illegal2:
//push	eax								//radix40_data[current_word_index] += weights[cycle] * j;
		mov		eax,1					//edx = weights[cycle]
		mul		edx								//eax:edx = weights[cycle] * j;
		add		[esi+ebx*4],eax					//radix40_data[current_word_index] += weights[cycle] * j;
mov		eax,1600					//edx = weights[cycle]
mul		edx								//eax:edx = weights[cycle] * j;
add		[esi+ebx*4+4],eax					//radix40_data[current_word_index] += weights[cycle] * j;
//pop		eax

mov edx,1		//eax % 3
mov eax,4
cmp esp,5
je do_single
cmp		esp,4
jbe		end_of_loop




		movzx	edx,byte ptr [edi+4]			//edx = ascii_data
movzx	ecx,byte ptr [edi+5]			//edx = ascii_data

movzx eax,HYPHEN
		movzx	edx,byte ptr [lookup_chars+edx]	//j = lookup_chars[ascii_data[i]];
movzx	ecx,byte ptr [lookup_chars+ecx]	//j = lookup_chars[ascii_data[i]];
/*
		test	dl,ILLEGAL
//jz		not_illegal1
		
//movzx	edx,HYPHEN						//j = HYPHEN
//mov		dword ptr [result],1			//S_ILLEGAL was not working.
cmovz edx,eax
//cmovz [result],eax  //gets set to 2,  Any non-zero value indicates error
test	cl,ILLEGAL
cmovz ecx,eax
//cmovz [result],eax  //gets set to 2,  Any non-zero value indicates error
*/
mov	eax,edx
and dl,not(0x80)  // do a "hyphen"
or	eax,ecx
and cl,not(0x80)  // do a "hyphen"
test al,ILLEGAL
//GET RID OF THE JUMP BY DOING THIS, and AL,ILLEGAL, OR DWORD PTR [result],EAX
jz not_illegal3
mov		dword ptr [result],1			//S_ILLEGAL was not working.
not_illegal3:
//push	eax								//radix40_data[current_word_index] += weights[cycle] * j;
		mov		eax,40					//edx = weights[cycle]
		mul		edx								//eax:edx = weights[cycle] * j;
		add		[esi+ebx*4+4],eax					//radix40_data[current_word_index] += weights[cycle] * j;
mov		eax,1					//edx = weights[cycle]
mul		edx								//eax:edx = weights[cycle] * j;
add		[esi+ebx*4+4],eax					//radix40_data[current_word_index] += weights[cycle] * j;
//pop		eax

//mov edx,1		//eax % 3
//mov eax,4
//cmp esp,1
//je do_single
//cmp		esp,4
//jbe		end_of_loop






		inc		ebx								//current_word_index++
		add		edi,6							//point to next array location in the ASCII array.

		sub		esp,6
		jnz		top_of_loop

jmp		end_of_loop

do_single:
		movzx	edx,byte ptr [edi+eax]			//edx = ascii_data, eax points to the offset in the buffer
		movzx	edx,byte ptr [lookup_chars+edx]	//j = lookup_chars[ascii_data[i]];
		test	dl,ILLEGAL
		jz		not_illegal4
		
		movzx	edx,HYPHEN						//j = HYPHEN
		mov		dword ptr [result],1			//S_ILLEGAL was not working.
not_illegal4:
//push	eax								//radix40_data[current_word_index] += weights[cycle] * j;
		mov		eax,[weights+edx*4]					//eax = weights[cycle], edx holds offset into buffer
		mul		edx								//eax:edx = weights[cycle] * j;
		add		[esi+ebx*4],eax					//radix40_data[current_word_index] += weights[cycle] * j;
//pop		eax

end_of_loop:
movd esp,mm0		//you cannot push and pop if this is not pointing to the right location.

		pop edi
		pop esi
		pop ebx

	}
	return(result);
}
#endif // #if USE_ALGO==ALGO3

//best time ? cycles
#if USE_ALGO==ALGO4
__forceinline int ascii_to_radix40_mark(Radix40 *radix40_data, unsigned char *ascii_data, int max_chars)
/* this routine converts a null-terminated ascii character string */
/* to radix 40 representation.  The allowable characters are: */
/* A-Z, 0-9, period, comma, hyphen, and space.  If any illegal characters */
/* are detected, they will be converted to hyphens, and the return value */
/* will be S_ILLEGAL. */
/* Lowercase letters will be upper-cased without error indication. */
/* The radix40 value will be padded with blanks to a multiple of three */
/* characters. If the number of characters in the ascii string is > max_chars */
/* only max_chars will be converted, and the return value will be S_ILLEGAL. */
/* If no error is detected, the return value will be S_OKAY. */
{
	int i;
	unsigned char j;
	int ascii_length;
	int result;
	int words_to_convert;
	int words_to_clear;
	int cycle;
	unsigned current_word_index;

	result = S_OKAY;
	ascii_length = strlen((char *)ascii_data);
	if (ascii_length > max_chars)
		{
		ascii_length = max_chars;
		result = S_ILLEGAL;
		}

	words_to_convert = ascii_length / 3;
	if (ascii_length % 3 != 0)
		words_to_convert ++;

	words_to_clear = max_chars / 3;
	if (max_chars % 3 != 0)
		words_to_clear ++;

   _asm {
      push ebx
      push esi
      push edi
//   memset(radix40_data,0,words_to_clear*sizeof(Radix40));
      mov      ecx,[words_to_clear]         //Get number of bytes to clear.
      xor      eax,eax
      mov      edi,[radix40_data]            //load up EDI with offset
      rep      stosd

//edx now free since we unrolled it 3 times.
      mov      eax,[ascii_length]            //eax = i = ascii_length ( we are counting down instead of up)
      mov      ebx,-1                     //current_word_index = ebx = -1
      mov      esi,[radix40_data]
      mov      edi,[ascii_data]
align 4
top_of_loop:
      inc      ebx                        //current_word_index++

      movzx   edx,byte ptr [edi+0]         //edx = ascii_data
      movzx   edx,byte ptr [lookup_chars+edx]   //j = lookup_chars[ascii_data[i]];
      test   dl,ILLEGAL
      jz      not_illegal1

      movzx   edx,HYPHEN                  //j = HYPHEN
      mov      dword ptr [result],1         //S_ILLEGAL was not working.
not_illegal1:
      mov      edx,table1600[edx*4]
      add      [esi+ebx*4],edx               //radix40_data[current_word_index] += weights[cycle] * j;

      cmp      eax,1
      jbe      end_of_loop

      movzx   edx,byte ptr [edi+1]         //edx = ascii_data
      movzx   edx,byte ptr [lookup_chars+edx]   //j = lookup_chars[ascii_data[i]];
      test   dl,ILLEGAL
      jz      not_illegal2

      movzx   edx,HYPHEN                  //j = HYPHEN
      mov      dword ptr [result],1         //S_ILLEGAL was not working.
not_illegal2:
      mov      edx,table40[edx*4]
      add      [esi+ebx*4],edx               //radix40_data[current_word_index] += weights[cycle] * j;

      cmp      eax,2
      jbe      end_of_loop

      movzx   edx,byte ptr [edi+2]         //edx = ascii_data
      movzx   edx,byte ptr [lookup_chars+edx]   //j = lookup_chars[ascii_data[i]];
      test   dl,ILLEGAL
      jz      not_illegal3

      movzx   edx,HYPHEN                  //j = HYPHEN
      mov      dword ptr [result],1         //S_ILLEGAL was not working.
not_illegal3:
//      mov      edx,table1[edx*4]
      add      [esi+ebx*4],edx               //radix40_data[current_word_index] += weights[cycle] * j;

      add      edi,3                     //point to next array location in the ASCII array.

      sub      eax,3
      jnz      top_of_loop

end_of_loop:

      pop edi
      pop esi
      pop ebx
   }


	return(result);
}
#endif // #if USE_ALGO==ALGO4

#endif // #if USE_ASM